/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "Channel4ProtocolCrypt.h"
#include <stdio.h>
#include <stdlib.h>
#include <boost/format.hpp>

// ##################################################################
/**
* @brief Initialisieren der Membervariablen.
*
* @param channelID Nummer der Kanalklasse fuer z.B. Trace.
* @param deviceID Busnummer der Kanalklasse.
*/
Channel4ProtocolCrypt::Channel4ProtocolCrypt( short channelId, short deviceId ) : Channel4Protocol( channelId, deviceId ) 
{
    // Umgebung zuruecksetzen
    reset();

    // Zufallsgenerator starten
    srand( (unsigned int)time( NULL ) );

    // Halten der vorhergehenden Sendedaten, um bei Wiederholdungen diese wiederzuverwenden.
    _packetNumberPrevious = 0xFF;
    memset( _packetDataPrevious, 0, sizeof(_packetDataPrevious) );
    _packetDataSizePrevious = 0;
}

// ##################################################################
/**
* @brief Destruktor.
*
*/
Channel4ProtocolCrypt::~Channel4ProtocolCrypt() 
{
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void Channel4ProtocolCrypt::getErrorModule( short module, short error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s( %i, %i, %p )\n", __FILE__, __LINE__, __func__, module, error, text.c_str() );

    // Ist IV_MODULE_PROTOCOL_CRYPT ?
    if ( module == IV_MODULE_PROTOCOL_CRYPT )
    {

    }

    // Ermittlung des Fehlertextes an Basisklasse abgeben.
    Channel4Protocol::getErrorModule( module, error, text );
}


// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void Channel4ProtocolCrypt::getErrorFunction( short module, short func, short error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s( %i, %i, %i, %p )\n", __FILE__, __LINE__, __func__, module, func, error, text.c_str() );

    if ( module == IV_MODULE_PROTOCOL_CRYPT )
    {
        // Sonderfehler für Passwortschutz abhandeln.
        if ( error == 28 )
        {
            // Die Modulfunktionen liefern den Fehlercode 28, wenn kein oder ein falsches Passwort verwendet wurde.
            // Es ist in diesem Fall zuerst das richtige Passwort für die Kommunikation erforderlich.
            text = boost::str( boost::format( "( %i ): Korrektes Passwort für die Kommunikation erforderlich." ) % error );
            return;
        }
        else if ( error == 29 )
        {
            // Die Modulfunktionen liefern den Fehlercode 29, wenn 3 mal ein falsches Passwort verwendet wurde.
            // Die Befehlsausführung ist danach für 60 Sekunden blockiert.
            text = boost::str( boost::format( "( %i ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt." ) % error );
            return;
        }

        switch ( func )
        {
        case IV_FUNCTION_EXCHANGE_PUBLIC_KEY:
            text = boost::str( boost::format( "( %i ): Nicht eingetragener Funktionsfehler!" ) % error );
            break;

        default:
            text = boost::str( boost::format( "( %i ): Nicht eingetragener Fehlerblock für Funktion!" ) % error );
            break;
        }

        return;
    }

    // Ermittlung des Fehlertextes an Basisklasse abgeben.
    Channel4ProtocolCrypt::getErrorFunction( module, func, error, text );
}

//#########################################################
//####################### HELPER ##########################
//#########################################################

// ##################################################################
/**
* @brief Initialisiert die grundlegend benoetigten Membervariablen.
*/
void Channel4ProtocolCrypt::reset()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    _result = IV_ERROR_SUCCESS;
    resetError();
    memset( _packetData, 0, sizeof( _packetData ) );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param deviceIDRemote Anzusprechende Geraetenummer.
*/
// Modulbefehl senden und Antwort zurueckgeben
//  deviceIDRemote -> Adresse (Geraetenummer) des Empfaengers
//  module   -> Nummer des anzusprechenden Modules
//  func -> Nummer der Funktion im Module
//	buf	     -> Buffer fuer Funktionsparameter.
//	count    -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
//
//  return -> <0 -> Kommunikationsfehler
//             0 -> senden und empfangen und ausfuehren erfolgreich
//			  >0 -> module-/Funktionsfehler
int Channel4ProtocolCrypt::call( short context, short module, short func, unsigned char *buf, int &count, int contextValue )
{
    trace( "%s(%i):INFO(3): %s( %i, %i, %i, %p, %i, %i )\n", __FILE__, __LINE__, __func__, context, module, func, buf, count, contextValue );
    int repeat = 0;

    // Im Crypt-Mode Userpassword muss auch ein solches vorliegen!
    // Bei Ping-Befehl pruefung unterlassen.
    if ( (context == IV_CONTEXT_MASTER)
        && !((module == IV_MODULE_PROTOCOL) && (func == IV_FUNCTION_PING))
        && getCryptSession().m_cryptMode == IV_PACKET_CRYPT_MODE_PASSPHRASE && !getCryptSession().IsValidUserKey() )
    {
        return -IV_ERROR_NEED_USER_PASSWORD;
    }

label_again:
    // Muss eine Schluesselaushandlung aufgrund folgender Faelle durchgefuehrt werden:
    // Es ist eine Anfrage (ausser Schluesselaustausch selbst) die verschluesselte Kommunikation wird unterstuetzt und
    // soll eingesetzt werden und 
    // 1. Es wurde noch kein Session-Key ausgehandelt.
    // 2. Der aktuelle Session-Key ist aufgrund eines Timeouts ungueltig.
    if ( (context == IV_CONTEXT_MASTER)
        && !((module == IV_MODULE_PROTOCOL_CRYPT) && (func == IV_FUNCTION_EXCHANGE_PUBLIC_KEY)) && (getCryptSession().m_cryptMode > 1) 
        && ((getCryptSession().m_sessionTime == 0) || (time(NULL) > (getCryptSession().m_sessionTime + getCryptSession().m_sessionTimeout))) )
    {
        if ( getCryptSession().m_sessionTime )
        {
            trace( "%s(%i):WARNING(1):  %s() [ session t(time): %i, t(timeout): %i, t(time-timeout): %i ]\n", __FILE__, __LINE__, __func__, static_cast<int>( time( NULL ) - getCryptSession().m_sessionTime ), getCryptSession().m_sessionTimeout, static_cast<int>( time( NULL ) - (getCryptSession().m_sessionTime + getCryptSession().m_sessionTimeout) ) );
        }

        // Aushandlung bereits durchgefuehrt und laenger als 9 Sekunden abgelaufen?
        // Dann muessen die Crypt-Routinen neu aufgesetzt werden!
        if ( (getCryptSession().m_sessionTime > 0) && ((time( NULL ) - (getCryptSession().m_sessionTime + getCryptSession().m_sessionTimeout))) > 9 )
        {
            getCryptSession().m_sessionTime = 0; // Session abgelaufen.
            getCryptSession().InitCryptMode();
            trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %i, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, getCryptSession().m_cryptMode, (getCryptSession().m_sessionTime == 0) ? "no" : "yes" );
        }

        // Buffer fuer Folgebefehl zwischenspeichern,
        // da das hinter buf liegende Bytearray durch Aufruf von exchangeSessionKey veraendert wird.
        int lCount = count;
        unsigned char data[IV_PACKET_LENGTH_DATA];
        memset( data, 0, sizeof( data ) );
        memcpy( data, buf, count );

        // Aushandlung eines Session-Key durchfuehren
        int lRes = exchangeSessionKey( getDeviceIdRemote() );
        if ( lRes == 0x0D00 || lRes == -IV_ERROR_NEED_USER_PASSWORD )
            return -IV_ERROR_NEED_USER_PASSWORD;

        // Aushandlung erfolgreich durchgefuehrt, Parameterdaten wiederherstellen.
        count = lCount;
        memcpy( buf, data, count );
    }

    int res = Channel4Protocol::call( context, module, func, buf, count, contextValue );
    // PROT_MODULE_ERROR_DECRYPT                (0x0D00) // Entschluesselung des Packetinhaltes bring Fehler.
    //                                                      Ablauf des Session-Key, Neuaushandlung ist erforderlich.
    if ( res == 0x0D00 )
    {
        //if ( module == IV_MODULE_PROTOCOL_CRYPT && func == IV_FUNCTION_EXCHANGE_PUBLIC_KEY )
        //  return res;

        getCryptSession().m_sessionTime = 0; // Session-Key ungueltig.
        getCryptSession().InitCryptMode();
        trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %i, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, getCryptSession().m_cryptMode, (getCryptSession().m_sessionTime == 0) ? "no" : "yes" );
        if ( repeat == 0 )
        {
            repeat++;
            goto label_again;
        }
        else
        {
            // Das aktuell angegebene User-Passwort ist ungueltig.
            return -IV_ERROR_NEED_USER_PASSWORD;
        }
    }

    // PROT_MODULE_ERROR_CRYPTED                (0x0E00) // Mode passt nicht. Z. B. Geraet arbeitet verschluesselt und Anfrage war unverschluesselt.
    if ( res == 0x0E00 )
    {
        // Fehler bei Entschluesselung der Daten.
        getCryptSession().m_cryptMode = 0;
        getCryptSession().m_sessionTime = 0;
    }

    return res;
}

int Channel4ProtocolCrypt::writeProtocolCrypt()
{
    // Handelt es sich um die vorhergehende Packetnummer?
    bool resend = (_packetNumberPrevious == _packetWrite[IV_PACKET_LENGTH_HEAD]);
    // Packettyp uebernehmen
    //  unsigned char type = _packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3]];
    // Anzahl der Datenbyte
    int count = *reinterpret_cast<unsigned short *>( &_packetWrite[4] ) - IV_PACKET_LENGTH_HEAD - _packetWrite[3] - sizeof( unsigned char ) - sizeof( unsigned short ) * 2;

    // Aktuell intern verwendeten Crypt-Mode eintragen.
    _packetWrite[6] |= getCryptSession().m_cryptMode;

    // Wenn cryptMode > 1 (dann wird die Verschluesselung von der Gegenseite unterstuetzt),
    // dann verschluesseln und Verschluesselte Daten fuer ggf. wiederholtes
    // senden halten.
    if ( getCryptSession().m_cryptMode > 1 )
    {
        unsigned short size;
        // Daten werden verschluesselt.
        _packetWrite[6] |= IV_PACKET_FLAG_CRYPTED;

        // Wiederholtes Senden?
        if ( resend )
        {
            // Zwischengespeicherten Dateninhalt wiederholt senden.
            memcpy( &_packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3] + sizeof( unsigned char )], _packetDataPrevious, _packetDataSizePrevious );
            size = _packetDataSizePrevious;
        }
        else
        {
            size = getCryptSession().TransmitEncrypt( &_packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3] + sizeof( unsigned char )], sizeof( unsigned short ) * 2 + count );
            // Aktuelle Daten fuer ggf. wiederholtes Senden halten.
            memcpy( _packetDataPrevious, &_packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3] + sizeof( unsigned char )], size );
            _packetDataSizePrevious = size;
        }

        // Anzahl Bytes die fuer die Verschluesselung hinzugefuegt wurde (16Byte Bloecke) berechnen.
        size -= sizeof( unsigned short ) * 2 + count;
        // Auf die Gesamtgroesse des Pakets draufaddieren.
        *reinterpret_cast<unsigned short *>( &_packetWrite[4] ) += size;
        _packetNumberPrevious = _packetWrite[IV_PACKET_LENGTH_HEAD];
    }

    return IV_ERROR_SUCCESS;
}

int Channel4ProtocolCrypt::writeStreamCrypt( std::string &buffer )
{
    // Wenn cryptMode > 1 (dann wird die Verschluesselung von der Gegenseite unterstuetzt),
    // dann verschluesseln und Verschluesselte Daten fuer ggf. wiederholtes
    // senden halten.
    if ( getCryptSession().m_cryptMode > 1 )
    {
        unsigned int count_crypt_byte;
        unsigned int count_blocks;

        // Anzahl der Datenbyte
        int count = buffer.size();

        // Anzahl 16 Byte Bloecke berechnen und Streamlaenge entsprechend anpassen.
        // Zusaetzlich werden die Daten um einen CRC und Anzahl Nutzdaten erweitert, daher (count + 5)!
        count_blocks = (count + 5) / 16;
        if ((count + 5) % 16)
            count_blocks++;

        count_crypt_byte = count_blocks * 16;

        // Puffer auf benoetigte Laenge anpassen.
        // Die Routine fuer Crypt fuellt den letzen 16 Byte Block des Puffer ggf. mit Zufallswerten auf.
        buffer.resize( count_crypt_byte, ' ' );

        // Datenblock verschluesseln.
        getCryptSession().TransmitEncryptStream( reinterpret_cast<unsigned char*>( const_cast<char*>( buffer.data() ) ), count );
    }

    return IV_ERROR_SUCCESS;
}

int Channel4ProtocolCrypt::readProtocolCrypt()
{
    // Packettyp uebernehmen
    //  unsigned char type = _packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3]];
    // Anzahl der Datenbyte
    int count = *reinterpret_cast<unsigned short *>( &_packetRead[4] ) - IV_PACKET_LENGTH_FRAME - _packetRead[3] - sizeof( unsigned char ) - sizeof( unsigned short ) * 2;

    // Wenn cryptMode > 1 (dann wird die Verschluesselung unterstuetzt), 
    // zusaetzlich die Daten verschluesselt vorliegen, dann entschluesseln.
    if ( (getCryptSession().m_cryptMode > 1) && (_lastFlags & IV_PACKET_FLAG_CRYPTED) )
    {
        unsigned short size = 2 * sizeof( unsigned short ) + count;
        if ( !getCryptSession().ReceiveDecrypt( &_packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3] + sizeof( unsigned char )], &size ) )
            return 0x0D00;

        *reinterpret_cast<unsigned short *>( &_packetRead[4] ) -= count - (size - 2 * sizeof( unsigned short ));
        count = size - 2 * sizeof( unsigned short );
    }

    // Liegt noch keine Information ueber die unterstuetzte Verschluesselung der Gegenstelle vor,
    // muss diese ermittelt werden.
    if ( getCryptSession().m_cryptMode != (_lastFlags & 0x3) )
    {
        trace( "%s(%i):WARNING(1):  %s() [ cryptMode change %02X -> %02X ]\n", __FILE__, __LINE__, __func__, getCryptSession().m_cryptMode, (_lastFlags & 0x3) );

        // Die verschluesselte Kommunikation wird auf eine der folgenden Arten unterstuetzt.
        if ( (_lastFlags & 0x3) == IV_PACKET_CRYPT_MODE_NO
            || (_lastFlags & 0x3) == IV_PACKET_CRYPT_MODE_DEFAULT
            || (_lastFlags & 0x3) == IV_PACKET_CRYPT_MODE_PASSPHRASE )
        {
            // Uebernahme des Verschluesselungsmode
            getCryptSession().m_cryptMode = (_lastFlags & 0x3);
        }
        else
        {
            // Die verschluesselte Kommunikation wird nicht unterstuetzt,
            // der Verschluesselungsmode ist somit 0.
            getCryptSession().m_cryptMode = 0;
        }

        // Initialisieren des aktuell angenommenen Crypt-Mode.
        getCryptSession().m_sessionTime = 0; // Eventuell aktueller Session-Key ist ungueltig.
        getCryptSession().InitCryptMode();
        trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %i, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, getCryptSession().m_cryptMode, (getCryptSession().m_sessionTime == 0) ? "no" : "yes" );
    }

    return IV_ERROR_SUCCESS;
}

int Channel4ProtocolCrypt::readStreamCrypt( std::string &buffer )
{
    // Wenn cryptMode > 1 (dann wird die Verschluesselung unterstuetzt), 
    // zusaetzlich die Daten verschluesselt vorliegen, dann entschluesseln.
    if ( getCryptSession().m_cryptMode > 1 )
    {
        unsigned int size = buffer.size();
        if ( !getCryptSession().ReceiveDecryptStream( reinterpret_cast<unsigned char*>( const_cast<char*>( buffer.data() ) ), &size ) )
            return 0x0D00;

        buffer.resize( size );
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Austausch eines Session-Key fuer die verschluesselte Kommunikation.
*
* @param deviceID Anzusprechende Geraetenummer.
*/
int Channel4ProtocolCrypt::exchangeSessionKey( short deviceId )
{
    trace( "%s(%i):INFO(3):  %s( %i )\n", __FILE__, __LINE__, __func__, deviceId );

    int count = 0;
    _result = IV_ERROR_FAIL;

    // Funktion in Remoteklasse aufrufen?
    if ( getDeviceId() != deviceId )
    {
        // Umgebung zuruecksetzen
        reset();

        // Remoteadresse setzen.
        setDeviceIdRemote( deviceId );

        // Aufbau eines 8 Byte zufallsteils vom Session-Key
        unsigned char myKey[8];
        memset( myKey, 0, sizeof( myKey ) );
        int i;
        // Einen 8 Byte Langen Zufallswert erstellen.
        for ( i = 0; i < static_cast<int>( sizeof( myKey ) ); ++i )
            myKey[i] = rand() % 255;

        // Daten in Sendebuffer kopieren
        memcpy( _packetData, myKey, sizeof( myKey ) );	 
        count = sizeof(myKey);
        if ( getCryptSession().m_cryptMode == IV_PACKET_CRYPT_MODE_PASSPHRASE )
            *reinterpret_cast<unsigned int *>( &_packetData[count] ) = getCryptSession().m_sessionTimeout;
        else
            *reinterpret_cast<unsigned int *>( &_packetData[count] ) = 3600; // 1 Stunde default
        
        count += 4;

        if ( (_result = call( IV_CONTEXT_MASTER, IV_MODULE_PROTOCOL_CRYPT, IV_FUNCTION_EXCHANGE_PUBLIC_KEY, _packetData, count )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            traceError( deviceId, IV_MODULE_PROTOCOL_CRYPT, IV_FUNCTION_EXCHANGE_PUBLIC_KEY );
            return _result;
        }

        // Antwortparameter pruefen
        if ( count != 8 )
        {
            trace( "%s(%i):WARNING(1):  %s() [ Antwortparameter fehlt ]\n", __FILE__, __LINE__, __func__, count );
        }

        // Aktuellen Aushandlungszeitpunkt halten
        getCryptSession().m_sessionTime = time( NULL );

        // Gesamten Session-Key uebernehmen.
        memcpy( getCryptSession().m_sessionKey, myKey, 8 );
        // Uebernehmen des Session-Key Teils vom Geraet.
        memcpy( &getCryptSession().m_sessionKey[8], _packetData, 8 );
        getCryptSession().m_sessionKeySize = 16;

        // Schluesselaushandlung erfolgreich.
        trace( "%s(%i):INFO(1): %s() [ Session-Key valid ]\n", __FILE__, __LINE__, __func__ );

        // Initialisieren der Verschluesselung (verwendet m_sessionTime, m_sessionKey)
        getCryptSession().InitCryptMode();
        trace( "%s(%i):INFO(1): %s() [ Crypt-Mode: %i, Valid Session-Key: %s ]\n", __FILE__, __LINE__, __func__, getCryptSession().m_cryptMode, (getCryptSession().m_sessionTime == 0) ? "no" : "yes" );
    }

    return _result;
}

int Channel4ProtocolCrypt::setCommunicationPassword( unsigned char *password, int size, int sessionTimeout, int global ) 
{
    return getCryptSession().SetCommunicationPassword( password, size, sessionTimeout, global );
}
